/*
make
./lesson05
*/

/*
 * This code was created by Jeff Molofee '99 
 * (ported to Linux/SDL by Ti Leggett '01)
 *
 * If you've found this code useful, please let me know.
 *
 * Visit Jeff at http://nehe.gamedev.net/
 * 
 * or for port-specific comments, questions, bugreports etc. 
 * email to leggett@eecs.tulane.edu
 */
 
 
/*****************************************************
SMLPRT
Small Particle Gravity Simulator
by Chris Merck <navaburo@gmail.com>
Compiles best on Linux (but may compile on other systems)

   QBasic Build: March 2003
   C Build: May 2003
   Linux C build(s): April 2006

This software may be freely redistributed under the terms
 of the GNU public license.

 png code by Guillaume Cottenceau
 
 base SDL code taken from:
 Andrew M. <andrew@textux.com>
*****************************************************/ 

 
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <time.h>
#include <SDL.h>
#include <GL/gl.h>
#include <GL/glu.h>

/* screen width, height, and bit depth */
#define SCREEN_WIDTH  800
#define SCREEN_HEIGHT 800
#define SCREEN_BPP     16

/* Define our booleans */
#define TRUE  1
#define FALSE 0





#define FONT_HEIGHT  8
#define FONT_WIDTH   8

/*define edges of screen*/
#define XMAX 640
#define YMAX 480
#define ZMAX 480

#define  NUMPRT 1000

#define  TAILLEN 15


#define FPS_DISP_TIMER_MAX 100

#define DISTMODMIN .6


#define FDIV 10000

//SFACTOR is the global energy retention per frame
#define SFACTOR .99
//RUBFACTOR is the energy retention when within distmod of another particle
#define RUBFACTOR .9999
#define RUBDIST 50

#define WALLRUBFACTOR .99

//this pulls the particles down uniformly
	#define GRAVITY .01

#define INTERATTRACT_CUTOFF_FACTOR 2

#define DEFAULT_FUNCTION 3

#define walkspd 1

/* This is our SDL surface */
SDL_Surface *surface;

GLfloat camx=0, camy=0, camz=0, camlookx=0, camlooky=0, camlookz=100;

void setpixel(SDL_Surface *screen, int x, int y, Uint8 r, Uint8 g, Uint8 b)
{
  Uint8 *ubuff8;
  Uint16 *ubuff16;
  Uint32 *ubuff32;
  Uint32 color;
  char c1, c2, c3;
  
  /* Lock the screen, if needed */
  if(SDL_MUSTLOCK(screen)) {
    if(SDL_LockSurface(screen) < 0) 
      return;
  }
  
  /* Get the color */
  color = SDL_MapRGB( screen->format, r, g, b );
  
  /* How we draw the pixel depends on the bitdepth */
  switch(screen->format->BytesPerPixel) 
    {
    case 1: 
      ubuff8 = (Uint8*) screen->pixels;
      ubuff8 += (y * screen->pitch) + x; 
      *ubuff8 = (Uint8) color;
      break;

    case 2:
      ubuff8 = (Uint8*) screen->pixels;
      ubuff8 += (y * screen->pitch) + (x*2);
      ubuff16 = (Uint16*) ubuff8;
      *ubuff16 = (Uint16) color; 
      break;  

    case 3:
      ubuff8 = (Uint8*) screen->pixels;
      ubuff8 += (y * screen->pitch) + (x*3);
      

      if(SDL_BYTEORDER == SDL_LIL_ENDIAN) {
	c1 = (color & 0xFF0000) >> 16;
	c2 = (color & 0x00FF00) >> 8;
	c3 = (color & 0x0000FF);
      } else {
	c3 = (color & 0xFF0000) >> 16;
	c2 = (color & 0x00FF00) >> 8;
	c1 = (color & 0x0000FF);	
      }

      ubuff8[0] = c3;
      ubuff8[1] = c2;
      ubuff8[2] = c1;
      break;
      
    case 4:
      ubuff8 = (Uint8*) screen->pixels;
      ubuff8 += (y*screen->pitch) + (x*4);
      ubuff32 = (Uint32*)ubuff8;
      *ubuff32 = color;
      break;
      
    default:
      fprintf(stderr, "Error: Unknown bitdepth!\n");
    }
  
  /* Unlock the screen if needed */
  if(SDL_MUSTLOCK(screen)) {
    SDL_UnlockSurface(screen);
  }
}

void iniprt(double *x, double *y, double *z)
{
  double tmp1, tmp2;

   /* Erase old pixel */
   //vga_setcolor(0);
   //vga_drawpixel(*x,*y);

   /* Initialize/reset the particle position */
   
   //this tmp crap is unnessisary. but HOW DO I CONVERT int to double otherwise???
   tmp1=rand();
   tmp2=RAND_MAX;
   tmp1=tmp1/tmp2;
   tmp1=200*tmp1;
   *x=tmp1;
   
   tmp1=rand();
   tmp2=RAND_MAX;
   tmp1=tmp1/tmp2;
   tmp1=200*tmp1;
   *y=tmp1;
   
   tmp1=rand();
   tmp2=RAND_MAX;
   tmp1=tmp1/tmp2;
   tmp1=200*tmp1;
   *z=tmp1;
}



void drawPart(SDL_Surface *screen, int x, int y, int z, Uint8 r, Uint8 g, Uint8 b)
{
    /* rotational vars for the triangle and quad, respectively */
    static GLfloat rtri, rquad;
    
    /* Move Left 1.5 Units And Into The Screen 6.0 */
    glLoadIdentity();
    glTranslatef( x/10.0f - 30.0f, -y/10.0f + 25.0f, -60.0f - z/10.0f);

    /* Rotate The Triangle On The Y axis ( NEW ) */
    glRotatef( rtri, 0.0f, 1.0f, 0.0f );

    glBegin( GL_TRIANGLES );             /* Drawing Using Triangles       */
      glColor3f(   1.0f,  0.0f,  0.0f ); /* Red                           */
      glVertex3f(  0.0f,  1.0f,  0.0f ); /* Top Of Triangle (Front)       */
      glColor3f(   0.0f,  1.0f,  0.0f ); /* Green                         */
      glVertex3f( -1.0f, -1.0f,  1.0f ); /* Left Of Triangle (Front)      */
      glColor3f(   0.0f,  0.0f,  1.0f ); /* Blue                          */
      glVertex3f(  1.0f, -1.0f,  1.0f ); /* Right Of Triangle (Front)     */

      glColor3f(   1.0f,  0.0f,  0.0f ); /* Red                           */
      glVertex3f(  0.0f,  1.0f,  0.0f ); /* Top Of Triangle (Right)       */
      glColor3f(   0.0f,  0.0f,  1.0f ); /* Blue                          */
      glVertex3f(  1.0f, -1.0f,  1.0f ); /* Left Of Triangle (Right)      */
      glColor3f(   0.0f,  1.0f,  0.0f ); /* Green                         */
      glVertex3f(  1.0f, -1.0f, -1.0f ); /* Right Of Triangle (Right)     */

      glColor3f(   1.0f,  0.0f,  0.0f ); /* Red                           */
      glVertex3f(  0.0f,  1.0f,  0.0f ); /* Top Of Triangle (Back)        */
      glColor3f(   0.0f,  1.0f,  0.0f ); /* Green                         */
      glVertex3f(  1.0f, -1.0f, -1.0f ); /* Left Of Triangle (Back)       */
      glColor3f(   0.0f,  0.0f,  1.0f ); /* Blue                          */
      glVertex3f( -1.0f, -1.0f, -1.0f ); /* Right Of Triangle (Back)      */

      glColor3f(   1.0f,  0.0f,  0.0f ); /* Red                           */
      glVertex3f(  0.0f,  1.0f,  0.0f ); /* Top Of Triangle (Left)        */
      glColor3f(   0.0f,  0.0f,  1.0f ); /* Blue                          */
      glVertex3f( -1.0f, -1.0f, -1.0f ); /* Left Of Triangle (Left)       */
      glColor3f(   0.0f,  1.0f,  0.0f ); /* Green                         */
      glVertex3f( -1.0f, -1.0f,  1.0f ); /* Right Of Triangle (Left)      */
    glEnd( );                            /* Finished Drawing The Triangle */

    /* Increase The Rotation Variable For The Triangle ( NEW ) */
    rtri  += 0.01f;
}

/* function to release/destroy our resources and restoring the old desktop */
void Quit( int returnCode )
{
    /* clean up the window */
    SDL_Quit( );

    /* and exit appropriately */
    exit( returnCode );
}

/* function to reset our viewport after a window resize */
int resizeWindow( int width, int height )
{
    /* Height / width ration */
    GLfloat ratio;
 
    /* Protect against a divide by zero */
   if ( height == 0 )
	height = 1;

    ratio = ( GLfloat )width / ( GLfloat )height;

    /* Setup our viewport. */
    glViewport( 0, 0, ( GLsizei )width, ( GLsizei )height );

    /* change to the projection matrix and set our viewing volume. */
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );

    /* Set our perspective */
    gluPerspective( 50.0f, ratio, 0.1f, 200.0f );
    
        /* Make sure we're chaning the model view and not the projection */
    glMatrixMode( GL_MODELVIEW );

    /* Reset The View */
    glLoadIdentity( );

    return( TRUE );
}

/* function to handle key press events */
void handleKeyPress( SDL_keysym *keysym )
{
    switch ( keysym->sym )
	{
	case SDLK_ESCAPE:
	    /* ESC key was pressed */
	    Quit( 0 );
	    break;
	case SDLK_F1:
	    /* F1 key was pressed
	     * this toggles fullscreen mode
	     */
	    SDL_WM_ToggleFullScreen( surface );
	    break;
	case SDLK_w: camz-=walkspd; printf("W %d\n", (int)camz); break;
	case SDLK_a: camx-=walkspd; break;
	case SDLK_s: camz+=walkspd; break;
	case SDLK_d: camx+=walkspd; break;
	default:
	    break;
	}

    return;
}

/* general OpenGL initialization function */
int initGL( GLvoid )
{

    /* Enable smooth shading */
    glShadeModel( GL_SMOOTH );

    /* Set the background black */
    glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );

    /* Depth buffer setup */
    glClearDepth( 1.0f );

    /* Enables Depth Testing */
    glEnable( GL_DEPTH_TEST );

    /* The Type Of Depth Test To Do */
    glDepthFunc( GL_LEQUAL );

    /* Really Nice Perspective Calculations */
    glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

    return( TRUE );
}

/* Here goes our frame drawing code */
int drawGLScene( GLvoid )
{

    /* These are to calculate our fps */
    static GLint T0     = 0;
    static GLint Frames = 0;
    /* Draw it to the screen */
    SDL_GL_SwapBuffers( );
    
    /* Clear The Screen And The Depth Buffer */
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    
    // set up our camera
    gluLookAt (camx, camy, camz, camlookx, camlooky, camlookz, 0.0f, 0.0f, 0.0f);
   
    
    /* Gather our frames per second */
    Frames++;
    {
	GLint t = SDL_GetTicks();
	if (t - T0 >= 5000) {
	    GLfloat seconds = (t - T0) / 1000.0;
	    GLfloat fps = Frames / seconds;
	    printf("%d frames in %g seconds = %g FPS\n", Frames, seconds, fps);
	    T0 = t;
	    Frames = 0;
	}
    }

    return( TRUE );
}

int main( int argc, char **argv )
{


	struct timeval {
		time_t         tv_sec;        /* seconds */
		suseconds_t    tv_usec;  /* microseconds */
	};
	
	int taxidist;
	int mx, my;
	int r,g,b;
	double tp;
	int walls=1; //this controls the existance of walls
	double PWR=10;
	int dffunc=DEFAULT_FUNCTION;
	double SRAMP=.5;
	double SCALE=100;
	double SPDMAX=10;
 	int np,n,m;
	double distmod;
	double tmp;
   	double dist, f;
	double x[NUMPRT], y[NUMPRT], z[NUMPRT], xs[NUMPRT], ys[NUMPRT], zs[NUMPRT];
	int friends[NUMPRT];
	FILE *fpdebug;
	
	printf("Small Partical Gravity Simulator\nBy Chris Merck\n");
	printf("This build requires SDL.\n Contact navaburo@gmail.com for more info\n");
	
	/*do {
	printf("\nNumber of Particals(max=%d):",NUMPRT);
   	scanf("%d",&np);
        } while (np>NUMPRT);*/
	np=300; //TODO: make this a command-line argument

	//initialize variables from command-line arguments
	/*if (argc<6)
	{
		printf("usage: pix silenti visablei cmin cmax amin amax\n");
		return(1);
	}
	SILENTI = atoi(argv[1]);
	VISABLEI = atoi(argv[2]);
	cmin = atof(argv[3]);
	cmax = atof(argv[4]);
	amin = atof(argv[5]);
	amax = atof(argv[6]);
	*/	
	
	//clear debug file
	system("rm debug.out");
	
	//open debug dump file
	if ((fpdebug = fopen("debug.out", "a+")) == NULL)
 	{
		printf("Cannot open debug file. \n");
		return(1);
	}
	
	srand(1);  /* Fetch a set of fresh random numbers */ //TODO: get a time based seed
	
	
	
	
	
			
	
		printf("initializing particles...\n");
		
		for (n=1; n!=np; n++) 
		{
		iniprt(&x[n],&y[n],&z[n]); /* Initialize Particle Positions */
		//fprintf(fpdebug, "\ninix=%1.2f\n", x[n]);
		}
		
		printf("begining main loop...\n");
		
	




    /* Flags to pass to SDL_SetVideoMode */
    int videoFlags;
    /* main loop variable */
    int done = FALSE;
    /* used to collect events */
    SDL_Event event;
    /* this holds some info about our display */
    const SDL_VideoInfo *videoInfo;
    /* whether or not the window is active */
    int isActive = TRUE;

    /* initialize SDL */
    if ( SDL_Init( SDL_INIT_VIDEO ) < 0 )
	{
	    fprintf( stderr, "Video initialization failed: %s\n",
		     SDL_GetError( ) );
	    Quit( 1 );
	}

    /* Fetch the video info */
    videoInfo = SDL_GetVideoInfo( );

    if ( !videoInfo )
	{
	    fprintf( stderr, "Video query failed: %s\n",
		     SDL_GetError( ) );
	    Quit( 1 );
	}

    /* the flags to pass to SDL_SetVideoMode */
    videoFlags  = SDL_OPENGL;          /* Enable OpenGL in SDL */
    videoFlags |= SDL_GL_DOUBLEBUFFER; /* Enable double buffering */
    videoFlags |= SDL_HWPALETTE;       /* Store the palette in hardware */
    videoFlags |= SDL_RESIZABLE;       /* Enable window resizing */

    /* This checks to see if surfaces can be stored in memory */
    if ( videoInfo->hw_available )
	videoFlags |= SDL_HWSURFACE;
    else
	videoFlags |= SDL_SWSURFACE;

    /* This checks if hardware blits can be done */
    if ( videoInfo->blit_hw )
	videoFlags |= SDL_HWACCEL;

    /* Sets up OpenGL double buffering */
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

    /* get a SDL surface */
    surface = SDL_SetVideoMode( SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP,
				videoFlags );

    /* Verify there is a surface */
    if ( !surface )
	{
	    fprintf( stderr,  "Video mode set failed: %s\n", SDL_GetError( ) );
	    Quit( 1 );
	}

    /* initialize OpenGL */
    initGL( );

    /* resize the initial window */
    resizeWindow( SCREEN_WIDTH, SCREEN_HEIGHT );

    /* wait for events */ 
    while ( !done )
	{
	    /* handle the events in the queue */

	    while ( SDL_PollEvent( &event ) )
		{
		    switch( event.type )
			{
			case SDL_ACTIVEEVENT:
			    /* Something's happend with our focus
			     * If we lost focus or we are iconified, we
			     * shouldn't draw the screen
			     */
			    if ( event.active.gain == 0 )
				isActive = FALSE;
			    else
				isActive = TRUE;
			    break;			    
			case SDL_VIDEORESIZE:
			    /* handle resize event */
			    surface = SDL_SetVideoMode( event.resize.w,
							event.resize.h,
							16, videoFlags );
			    if ( !surface )
				{
				    fprintf( stderr, "Could not get a surface after resize: %s\n", SDL_GetError( ) );
				    Quit( 1 );
				}
			    resizeWindow( event.resize.w, event.resize.h );
			    break;
			case SDL_KEYDOWN:
			    /* handle key presses */
			    handleKeyPress( &event.key.keysym );
			    break;
			case SDL_QUIT:
			    /* handle quit requests */
			    done = TRUE;
			    break;
			default:
			    break;
			}
		}

	    /* draw the scene */
	    if ( isActive ){
			for (n=1;n!=np;n++) {
				for (m=1;m!=np;m++) {
					if (n==m) continue; /* Don't pull itself! */
			
					/* Calculate distance */
					dist = sqrt(pow((x[m] - x[n]),2) + pow((y[m]-y[n]),2) + pow((z[m]-z[n]),2));
	
					if (dist<INTERATTRACT_CUTOFF_FACTOR*SCALE)
					{
					
					
					/* Calculate force exerted */
					distmod=dist/SCALE; //distbig 
					if (distmod<DISTMODMIN) distmod=DISTMODMIN;
					switch (dffunc)
					{
					case 0:
						//regular gravitation
						f = 50 / pow(dist,3);
						break;
					case 1:
						//repel when close, attract when far
						f = ((2*distmod-2)/distmod+2) / pow(distmod,PWR)/FDIV;
						break;
					case 2:
						//sinesoidal field (strange)
						f = -cos(distmod)/FDIV/1000.;
						break;
					case 3:
						//double hump (crude)
						f = -cos(distmod)/pow(distmod,PWR)/FDIV;
						break;
					case 4:
						//revised double hump
						f=-(PWR/10./(pow(distmod,2)+1.)-1./(5.*pow(distmod,2) + .3)-1./(5.*(pow(distmod-1.,2)+.5)))/FDIV;
					}
					
					if (dist<100) friends[n]++;
					//if (dist<20) friends[n]++;
					
					/*if (dist<SCALE*SRAMP) {
					//rub-off speed
					xs[n]=(99.*xs[n] + 1.*xs[m])/100.;
					ys[n]=(99.*ys[n] + 1.*ys[m])/100.;
					
					//rub off heat
					xs[n]*=RUBFACTOR;
					ys[n]*=RUBFACTOR;
					}*/
							
					/* Influence object in correct direction */
					xs[n] += f*(x[m]-x[n]);
					ys[n] += f*(y[m]-y[n]);
					zs[n] += f*(z[m]-z[n]);
					}
					
				} /* next M */
		
				//exert uniform (planitary) gravitation downwards
				ys[n] += GRAVITY;
				
				//repel from mouse
				/*dist = sqrt(pow((mx - x[n]),2) + pow((my-y[n]),2));
				if (dist<80) {
					f = -2. / pow(dist,2);
					xs[n] += f*(mx-x[n]);
					ys[n] += f*(my-y[n]);
				}*/
				
				/* Check for off screen, and bounce if walls, respawn if no walls */
				if (y[n] < 0)
				{
					//if (walls == -1){
					//	iniprt(&x[n],&y[n]);
					//	xs[n]=0; ys[n]=0;
					//} else {
					//	ys[n]=abs(ys[n]);
					//	y[n]=.5;
					//}
					
				}
				if (x[n] > XMAX)
				{
					if (walls == -1){
						iniprt(&x[n],&y[n],&z[n]);
						xs[n]=0; ys[n]=0; zs[n]=0;
					} else {
						xs[n]=-abs(xs[n]);
						x[n]=XMAX-.5;
					}
				}	
				if ( x[n] < 0 )
				{
					if (walls == -1){
						iniprt(&x[n],&y[n],&z[n]);
						xs[n]=0; ys[n]=0; zs[n]=0;
					} else {
						xs[n]=abs(xs[n]);
						x[n]=.5;
					}
				}
				if (z[n] > ZMAX)
				{
					if (walls == -1){
						iniprt(&x[n],&y[n],&z[n]);
						xs[n]=0; ys[n]=0; zs[n]=0;
					} else {
						zs[n]=-abs(zs[n]);
						z[n]=XMAX-.5;
					}
				}	
				if ( z[n] < 0 )
				{
					if (walls == -1){
						iniprt(&x[n],&y[n],&z[n]);
						xs[n]=0; ys[n]=0; zs[n]=0;
					} else {
						zs[n]=abs(zs[n]);
						z[n]=.5;
					}
				}
				if (y[n] > YMAX)
				{
					y[n] = YMAX - 1; //place back on screen
					ys[n] = 0; //slow down
				}	
				
				//the floor gives a repulsive force
				dist = abs(YMAX - y[n] - 250);
				if (dist < 50 && x[n] < XMAX/2 && z[n] < 200) f = -.1;
				/*
				distmod=dist/SCALE;
				if (distmod<DISTMODMIN) distmod=DISTMODMIN;
				f = 20./pow(distmod,2);
				*/
				ys[n] += f; //apply floor force
				
		
				/* Save old values, move the part and DRAW THAT PIXEL!!! */
				/*ox[n][TAILLEN] = x[n];
				oy[n][TAILLEN]= y[n];*/
		
				/* Save tail data */
				/*for (m=TAILLEN;m>=0;m--) {
					ox[n][TAILLEN-m]=ox[n][TAILLEN-m-1];
					oy[n][TAILLEN-m]=oy[n][TAILLEN-m-1];
				}*/
	
				//cap particle speed
				/*if (xs[n]>SPDMAX) xs[n]=SPDMAX;
				if (xs[n]<-SPDMAX) xs[n]=-SPDMAX;
				if (ys[n]>SPDMAX) ys[n]=SPDMAX;
				if (ys[n]<-SPDMAX) ys[n]=-SPDMAX;*/
	
				//lose some speed...
				xs[n]*=SFACTOR;
				ys[n]*=SFACTOR;
				zs[n]*=SFACTOR;
				
				//erase old particle
				//if (x[n] > 0 && x[n] < screen->w && y[n] > 0 && y[n] < screen->h)
				//	setpixel(screen,x[n],y[n],255,255,255);
				
				//move particle
				x[n] += xs[n];
				y[n] += ys[n];
				z[n] += zs[n];
				
				/*
				r = friends[n]*20;
				if (r>255) r=255;
				g=255-r;
				b=(ys[n]+xs[n])*2;
				if (b>255) b=255;
				*/
				
				//draw new particle
				if (x[n] > 0 && x[n] < surface->w && y[n] > 0 && y[n] < surface->h)
					drawPart(surface,x[n],y[n],z[n],r,g,b);
					
	
				/*oxa[n] = (oxa[n] + xs[n] - oxs[n])/2.; //record old acceleration
				oya[n] = (oya[n] + ys[n] - oys[n])/2.;
				oxs[n] = xs[n]; //record old speeds
				oys[n] = ys[n];*/
				
				//printf ("%1.2f\n",oxa[n]);
				
				friends[n]=0;
			}
		
		/*
		//erase old mouse dot
		vga_setcolor(0);
		vga_drawpixel(mx,my);
		//get the mouse position
		mouse_update();
		mouse_getposition_6d(&mx, &my, NULL, &rx, NULL, NULL);
		//draw a dot as a mouse cursor
		vga_setcolor(3);
		vga_drawpixel(mx,my);
		*/
		
		
		/*
		//lastkey=vga_getkey();
		if (lastkey==119) SCALE++; //w
		if (lastkey==115) SCALE--; //s
		if (lastkey==97)  PWR--; //a
		if (lastkey==100) PWR++; //d
		//if (lastkey==99) vga_clear(); //c = refresh
		if (lastkey==98) walls=-walls; //b = toggle walls on/off
		*/
		
		}
		
		camlookx = (GLfloat)x[1]; camlooky = (GLfloat)y[1]; camlookz = (GLfloat)z[1]; //TODO REMOVE ME
		
		//now we are done with this frame, so display it
		drawGLScene();
		
	}

    /* clean ourselves up and exit */
    Quit( 0 );

    /* Should never get here */
    return( 0 );
}
